@@ -6,7 +6,7 @@ use Time::Local;
 use strict;
 use vars qw($VERSION);
 
-$VERSION = '0.03';
+$VERSION = '0.04';
 
 my %months_map = (
     'Jan' => 0, 'Feb' => 1, 'Mar' => 2,
@@ -15,6 +15,40 @@ my %months_map = (
     'Oct' => 9, 'Nov' =>10, 'Dec' =>11
 );
 
+# fast timelocal
+my $str2time_last_time;
+my $str2time_last_day;
+my $str2time_last_month;
+my $enable_year_decrement = 1; # year-increment algorithm: if in january, if december is seen, decrement
+                               # year
+# 0: sec, 1: min, 2: h, 3: day, 4: month, 5: year
+sub str2time($$$$$$$)
+{
+    my $GMT = pop @_;
+    my $day_secs = $_[2]*3600+$_[1]*60+$_[0];
+    if(defined $str2time_last_time) {
+        if( $_[3] == $str2time_last_day and
+            $_[4] == $str2time_last_month )
+        {
+            return $str2time_last_time + $day_secs;
+        }
+    }
+
+    my $time;
+    if($GMT) {
+        $time = timegm(@_);
+    }
+    else {
+        $time = timelocal(@_);
+    }
+
+    $str2time_last_time = $time - $day_secs;
+    $str2time_last_day = $_[3];
+    $str2time_last_month = $_[4];
+
+    return $time;
+}
+
 sub new($$;%)
 {
     my ($class, $file, %data) = @_;
@@ -55,7 +89,7 @@ sub next($)
 
     while($self->{_repeat}>0) {
         $self->{_repeat}--;
-        return $self->{_last_data};
+        return $self->{_repeat_data};
     }
 
     line: while(my $str = $self->_next_line()) {
@@ -78,38 +112,43 @@ sub next($)
         defined $mon or croak "unknown month $1\n";
 
         # year change
-        $self->{year}++ if defined $self->{_last_mon}
-            and $self->{_last_mon} > $mon;
+        if($mon==0) {
+            $self->{year}++ if defined $self->{_last_mon} and $self->{_last_mon} == 11;
+            $enable_year_decrement = 1;
+        }
+        elsif($mon == 11) {
+            if($enable_year_decrement) {
+                $self->{year}-- if defined $self->{_last_mon} and $self->{_last_mon} != 11;
+            }
+        }
+        else {
+            $enable_year_decrement = 0;
+        }
+
         $self->{_last_mon} = $mon;
 
         # convert to unix time
-        my $time;
-        if($self->{GMT}) {
-            $time = timegm($5,$4,$3,$2,$mon,$self->{year}-1900);
-        }
-        else {
-            $time = timelocal($5,$4,$3,$2,$mon,$self->{year}-1900);
-        }
+        my $time = str2time($5,$4,$3,$2,$mon,$self->{year}-1900,$self->{GMT});
 
         my ($host, $text) = ($6, $7);
-        
+
         # last message repeated ... times
         if($text =~ /^last message repeated (\d+) time/) {
             next line if defined $self->{repeat} and not $self->{repeat};
-            next line if not defined $self->{_last_data};
+            next line if not defined $self->{_last_data}{$host};
             $1 > 0 or do {
                 carp "last message repeated 0 or less times??";
                 next line;
             };
             $self->{_repeat}=$1-1;
-            return $self->{_last_data};
+            $self->{_repeat_data}=$self->{_last_data}{$host};
+            return $self->{_last_data}{$host};
         }
 
         # some systems send over the network their
         # hostname prefixed to the text. strip that.
         $text =~ s/^[-\w+\.]+\s+//;
 
-        # snort log
         $text =~ /^
             (\S+?)          # program   -- 1
             (?:\[(\d+)\])?  # PID       -- 2
@@ -122,18 +161,29 @@ sub next($)
             next line;
         };
 
-        $self->{_last_data} = {
-            timestamp => $time,
-            host      => $host,
-            program   => $1,
-            pid       => $2,
-            msgid     => $3,
-            facility  => $4,
-            level     => $5,
-            text      => $6,
-        };
+        if($self->{arrayref}) {
+            $self->{_last_data}{$host} = [
+                $time,  # 0: timestamp 
+                $host,  # 1: host      
+                $1,     # 2: program   
+                $2,     # 3: pid       
+                $6,     # 4: text      
+                ];
+        }
+        else {
+            $self->{_last_data}{$host} = {
+                timestamp => $time,
+                host      => $host,
+                program   => $1,
+                pid       => $2,
+                msgid     => $3,
+                facility  => $4,
+                level     => $5,
+                text      => $6,
+            };
+        }
 
-        return $self->{_last_data};
+        return $self->{_last_data}{$host};
     }
     return undef;
 }
@@ -163,7 +213,7 @@ Parse::Syslog presents a simple interface to parse syslog files: you create
 a parser on a file (with B<new>) and call B<next> to get one line at a time
 with Unix-timestamp, host, program, pid and text returned in a hash-reference.
 
-=head1 Constructing a Parser
+=head2 Constructing a Parser
 
 B<new> requires as first argument a file-name for the syslog-file to be parsed.
 Alternatively, you can pass a File::Tail object as first argument, in which
@@ -191,9 +241,38 @@ Parse::Syslog will by default repeat xx times events that are followed by
 messages like 'last message repeated xx times'. If you set this option to
 false, it won't do that.
 
+=item B<arrayref>
+
+If this option is true, I<next> will return an array-ref instead of a hash-ref
+(and is thus a bit faster), with the following contents:
+
+=over 4
+
+=item 0:
+
+timestamp
+
+=item 1:
+
+host
+
+=item 2:
+
+program
+
+=item 3:
+
+pid
+
+=item 4:
+
+text
+
 =back
 
-=head1 Parsing the file
+=back
+
+=head2 Parsing the file
 
 The file is parse one line at a time by calling the B<next> method, which returns
 a hash-reference containing the following keys:
@@ -238,7 +317,7 @@ generation" enabled".
 
 =back
 
-=head1 BUGS
+=head2 BUGS
 
 There are many small differences in the syslog syntax between operating
 systems. This module has been tested for syslog files produced by the following
@@ -255,6 +334,8 @@ Report problems for these and other operating systems to the author.
 Copyright (c) 2001, Swiss Federal Institute of Technology, Zurich.
 All Rights Reserved.
 
+=head1 LICENSE
+
 This module is free software; you can redistribute it and/or
 modify it under the same terms as Perl itself.
 
@@ -262,6 +343,13 @@ modify it under the same terms as Perl itself.
 
 David Schweikert <dws@ee.ethz.ch>
 
+=head1 HISTORY
+
+ 2001-08-12 ds 0.01 first version
+ 2001-08-19 ds 0.02 fix 'last message repeated xx times', Solaris 8 problems
+ 2001-08-20 ds 0.03 implemented GMT option, year specification, File::Tail
+ 2001-10-31 ds 0.04 faster time parsing, implemented 'arrayref' option, better time-increment algorithm
+
 =cut
 
 # vi: sw=4 et
